import 'dart:io';
import 'dart:async';
import 'package:wrapper_args/wrapper_args.dart';
import '../utils/version.dart';

Command init() {
  final parser = ArgParser()
    ..addFlag(
      "help",
      abbr: "h",
      negatable: false,
      help: "Print this usage information",
    );
  return Command(
    name: "go",
    help: "generate go code",
    longHelp: "Usage: generate go [<subcommand>] [<options>]",
    parser: parser,
    execute: (command, results) {
      if (results["help"]) {
        print(command.usage);
      }
    },
  )..addCommand(
      [
        _init(),
        _version(),
      ],
    );
}

Command _version() {
  final parser = ArgParser()
    ..addFlag(
      "help",
      abbr: "h",
      negatable: false,
      help: "Print this usage information",
    )
    ..addOption(
      "control",
      defaultsTo: Version.defaultsTo,
      allowed: Version.allowed,
      help: "version control",
    )
    ..addOption(
      "output",
      abbr: "o",
      defaultsTo: "version/version.go",
    );
  return Command(
    name: "version",
    help: "generate go version code from source",
    longHelp: "Usage: generate go version [<options>]",
    parser: parser,
    execute: (command, results) {
      if (results["help"]) {
        print("${command.usage}");
        return 0;
      }
      () async {
        final stat = await FileStat.stat("go.mod");
        if (stat.type == FileSystemEntityType.notFound) {
          throw StateError("go.mod not found");
        }
        final output = results["output"];
        final version = Version.fromControl(results["control"]);

        print('''
tag      = ${version.tag} 
commit   = ${version.commit}
datetime = ${version.datetime}
''');

        var file = File(output);
        file = await file.create(recursive: false);
        await file.writeAsString('''
package version

// Tag .
const Tag = "${version.tag}";
// Commit .
const Commit = "${version.commit}";
// Date .
const Date = "${version.datetime}";
''');
      }();
    },
  );
}

Command _init() {
  final parser = ArgParser()
    ..addFlag(
      "help",
      abbr: "h",
      negatable: false,
      help: "Print this usage information",
    )
    ..addOption(
      "template",
      defaultsTo: "console",
      allowed: ["console"],
      allowedHelp: {
        "console": "A command-line application sample",
      },
      help: "project template",
    )
    ..addOption(
      "package-name",
      help: "package name",
    )
    ..addOption(
      "project-name",
      help: "package name",
    );
  return Command(
    name: "init",
    help: "init go project at this directory",
    longHelp: "Usage: generate go init [<options>]",
    parser: parser,
    execute: (command, results) {
      if (results["help"]) {
        print("${command.usage}");
        return 0;
      }
      final String packageName = results["package-name"] ?? "";
      if (packageName.isEmpty) {
        throw ArgumentError("--package-name not specified");
      }
      final String projectName = results["project-name"] ?? "";
      if (projectName.isEmpty) {
        throw ArgumentError("--project-name not specified");
      }
      final version = Version.fromControl(Version.defaultsTo);
      final goConsole = GoConsole(
        packageName: packageName,
        projectName: projectName,
        version: version,
      );

      goConsole.done();
    },
  );
}

String formatPackage(String name) {
  if (name == null) {
    throw UnsupportedError("package not support : null");
  }
  var str = name.trim();
  if (str.startsWith("/")) {
    str = str.substring(1);
  }
  if (str.endsWith("/")) {
    str = str.substring(
      0,
      str.length - 1,
    );
  }
  str = str.trim();
  if (str.isEmpty) {
    throw UnsupportedError("package not support : $name");
  }
  if (str.contains(r"//")) {
    throw UnsupportedError("package not support : $name");
  }
  if (!RegExp(r"^[0-9a-zA-Z][0-9a-zA-Z\_\/]*$").hasMatch(str)) {
    throw UnsupportedError("package not support : $name");
  }

  return str;
}

String formatProject(String name) {
  if (name == null) {
    throw UnsupportedError("project not support : null");
  }
  var str = name.trim();
  if (str.startsWith("/")) {
    str = str.substring(1);
  }
  if (str.endsWith("/")) {
    str = str.substring(
      0,
      str.length - 1,
    );
  }
  str = str.trim();
  if (str.isEmpty) {
    throw UnsupportedError("project not support : $name");
  }
  if (!RegExp(r"^[0-9a-zA-Z][0-9a-zA-Z\_]*$").hasMatch(str)) {
    throw UnsupportedError("project not support : $name");
  }
  return str;
}

class GoConsole {
  final String packageName;
  final String projectName;
  final Version version;
  GoConsole({String packageName, String projectName, this.version})
      : packageName = formatPackage(packageName),
        projectName = formatProject(projectName);

  Future<void> done() async {
    await _createGitignore();
    await _createMod();

    await _createVersion();
    await _createConfigure();
    await _createLogger();
    await _createUtils();

    await _createCommand();
    await _createMain();

    await _createSettings();
    await _createSettings();
    await _createTasks();
    await _createLaunch();
  }

  Future<void> _createLaunch() async {
    final file = File(".vscode/launch.json");
    if (await _existFile(file)) {
      return;
    }
    await file.create(recursive: true);
    await file.writeAsString('''
{
    // Use IntelliSense to learn about possible attributes.
    // Hover to view descriptions of existing attributes.
    // For more information, visit: https://go.microsoft.com/fwlink/?linkid=830387
    "version": "0.2.0",
    "configurations": [
        {
            "name": "Launch",
            "type": "go",
            "request": "launch",
            "mode": "exec",
            "program": "\${workspaceRoot}/bin/${projectName}d",
            "windows": {
                "program": "\${workspaceRoot}/bin/${projectName}d.exe",
            },
            "env": {},
            "args": [
                "-v"
            ]
        }
    ]
}
''');
  }

  Future<void> _createTasks() async {
    final file = File(".vscode/tasks.json");
    if (await _existFile(file)) {
      return;
    }
    await file.create(recursive: true);
    await file.writeAsString('''
{
    // See https://go.microsoft.com/fwlink/?LinkId=733558
    // for the documentation about the tasks.json format
    "version": "2.0.0",
    "tasks": [
        {
            "label": "build debug",
            "type": "shell",
            "command": "go build -o bin/${projectName}d",
            "windows": {
                "command": "go build -o bin/${projectName}d.exe",
            },
            "options": {
                "cwd": "\${workspaceRoot}"
            },
            "group": {
                "kind": "build",
                "isDefault": true
            },
            "problemMatcher": "\$go",
        },
        {
            "label": "build release",
            "type": "shell",
            "command": "go build -ldflags '-s -w' -o bin/$projectName",
            "windows": {
                "command": "go build -ldflags '-s -w' -o bin/$projectName.exe",
            },
            "options": {
                "cwd": "\${workspaceRoot}"
            },
            "group": {
                "kind": "build",
                "isDefault": true
            },
            "problemMatcher": "\$go",
        },
    ]
}
''');
  }

  Future<void> _createSettings() async {
    final file = File(".vscode/settings.json");
    if (await _existFile(file)) {
      return;
    }
    await file.create(recursive: true);
    await file.writeAsString('''
{
    "editor.formatOnSave": true,
    "files.exclude": {
        "**/.classpath": true,
        "**/.project": true,
        "**/.settings": true,
        "**/.factorypath": true,
        "go.sum": true,
    },
    "files.associations": {
        "*.json": "jsonc"
    },
}
''');
  }

  Future<void> _createUtils() async {
    final file = File("utils/func.go");
    if (await _existFile(file)) {
      return;
    }
    await file.create(recursive: true);
    await file.writeAsString('''
package utils

import (
	"log"
	"os"
	"os/exec"
	"path/filepath"
)

var _BasePath string

// BasePath return executable file directory
func BasePath() string {
	if _BasePath != "" {
		return _BasePath
	}
	filename, e := exec.LookPath(os.Args[0])
	if e != nil {
		log.Fatalln(e)
	}

	filename, e = filepath.Abs(filename)
	if e != nil {
		log.Fatalln(e)
	}
	_BasePath = filepath.Dir(filename)
	return _BasePath
}

// Abs Use bashPath as the working directory to return the absolute path
func Abs(bashPath, path string) string {
	if filepath.IsAbs(path) {
		path = filepath.Clean(path)
	} else {
		path = filepath.Clean(bashPath + "/" + path)
	}
	return path
}
''');
  }

  Future<void> _createLogger() async {
    final file = File("logger/init.go");
    if (await _existFile(file)) {
      return;
    }
    await file.create(recursive: true);
    await file.writeAsString('''
package logger

import (
	"log"
	"path/filepath"
	"strings"

	logger "gitlab.com/king011/king-go/log/logger.zap"
	"go.uber.org/zap"
)

// Logger logger single
var Logger logger.Logger

// JoinFields .
var JoinFields = logger.JoinFields

// Fields .
var Fields = logger.Fields

// Init init logger
func Init(basePath string, options *logger.Options) (e error) {
	// format
	options.Filename = strings.TrimSpace(options.Filename)
	if options.Filename != "" {
		if !filepath.IsAbs(options.Filename) {
			options.Filename = filepath.Clean(basePath + "/" + options.Filename)
		}
	}
	var zapOptions []zap.Option
	if options.Caller {
		zapOptions = append(zapOptions, zap.AddCaller())
	}

	// new logger
	l := logger.New(options, zapOptions...)

	// run http
	if options.HTTP != "" {
		errHTTP := l.StartHTTP()
		if errHTTP == nil {
			if l.OutFile() {
				log.Println("zap http running", options.HTTP)
			}
			l.Info("zap http",
				zap.String("running", options.HTTP),
			)
		} else {
			if l.OutFile() {
				log.Println("zap http running", errHTTP)
			}
			l.Warn("zap http",
				zap.Error(errHTTP),
			)
		}
	}
	// Attach
	Logger.Attach(l)
	return
}
''');
  }

  Future<void> _createConfigure() async {
    final file = File("configure/Configure.go");
    if (await _existFile(file)) {
      return;
    }
    await file.create(recursive: true);
    await file.writeAsString('''
package configure

import (
	"encoding/json"
	"io/ioutil"

	"github.com/google/go-jsonnet"
	logger "gitlab.com/king011/king-go/log/logger.zap"
)

// Configure .
type Configure struct {
	Logger logger.Options
}

// Format .
func (c *Configure) Format(basePath string) (e error) {

	return
}
func (c *Configure) String() string {
	if c == nil {
		return "nil"
	}
	b, e := json.MarshalIndent(c, "", "\t")
	if e != nil {
		return e.Error()
	}
	return string(b)
}

var _Configure Configure

// Single single Configure
func Single() *Configure {
	return &_Configure
}

// Load .
func (c *Configure) Load(filename string) (e error) {
	var b []byte
	b, e = ioutil.ReadFile(filename)
	if e != nil {
		return
	}
	vm := jsonnet.MakeVM()
	var jsonStr string
	jsonStr, e = vm.EvaluateSnippet("", string(b))
	if e != nil {
		return
	}
	b = []byte(jsonStr)
	e = json.Unmarshal(b, c)
	if e != nil {
		return
	}
	return
}
''');
  }

  Future<void> _createVersion() async {
    final file = File("version/version.go");
    if (await _existFile(file)) {
      return;
    }
    await file.create(recursive: true);
    await file.writeAsString('''
package version

// Tag .
const Tag = "${version.tag}";
// Commit .
const Commit = "${version.commit}";
// Date .
const Date = "${version.datetime}";
''');
  }

  Future<void> _createCommand() async {
    final file = File("cmd/init.go");
    if (await _existFile(file)) {
      return;
    }
    await file.create(recursive: true);
    await file.writeAsString('''
package cmd

import (
	"fmt"
	"log"
	"runtime"

	"$packageName/configure"
	"$packageName/logger"
	"$packageName/utils"
	"$packageName/version"

	"github.com/spf13/cobra"
)

const (
	// App .
	App = "$projectName"
)

var v bool
var rootCmd = &cobra.Command{
	Use:   App,
	Short: "generate tools create $projectName",
	Run: func(cmd *cobra.Command, args []string) {
		if v {
			fmt.Println(runtime.GOOS, runtime.GOARCH, runtime.Version())
			fmt.Println(version.Tag)
			fmt.Println(version.Commit)
			fmt.Println(version.Date)
		} else {
			fmt.Println(App)
			fmt.Println(runtime.GOOS, runtime.GOARCH, runtime.Version())
			fmt.Println(version.Tag)
			fmt.Println(version.Commit)
			fmt.Println(version.Date)
			fmt.Printf(`Use "%v --help" for more information about this program.
`, App)
		}
	},
}

func init() {
	flags := rootCmd.Flags()
	flags.BoolVarP(&v,
		"version",
		"v",
		false,
		"show version",
	)
	initTest()
}
func initTest() {
	var filename string
	basePaht := utils.BasePath()
	cmd := &cobra.Command{
		Use:   "test-init",
		Short: "test init",
		Run: func(cmd *cobra.Command, args []string) {
			// load configure
			cnf := configure.Single()
			e := cnf.Load(filename)
			if e != nil {
				log.Fatalln(e)
			}
			e = cnf.Format(basePaht)
			if e != nil {
				log.Fatalln(e)
			}

			// init logger
			e = logger.Init(basePaht, &cnf.Logger)
			if e != nil {
				log.Fatalln(e)
			}

			// done
			logger.Logger.Info("ok")
		},
	}
	flags := cmd.Flags()
	flags.StringVarP(&filename, "config",
		"c",
		utils.Abs(basePaht,"$projectName.jsonnet"),
		"configure file",
	)
	rootCmd.AddCommand(cmd)
}
// Execute run command
func Execute() error {
	return rootCmd.Execute()
}
''');
  }

  Future<void> _createMod() async {
    final file = File("go.mod");
    if (await _existFile(file)) {
      return;
    }
    await file.writeAsString('''
module $packageName
''');
  }

  Future<void> _createGitignore() async {
    final file = File(".gitignore");
    if (await _existFile(file)) {
      return;
    }
    await file.writeAsString('''
# Binaries for programs and plugins
*.exe
*.exe~
*.dll
*.so
*.dylib

# Test binary, build with `go test -c`
*.test

# Output of the go coverage tool, specifically when used with LiteIDE
*.out

/$projectName
/${projectName}d
/bin/$projectName
/bin/${projectName}d
*.tar.gz
''');
  }

  Future<void> _createMain() async {
    final file = File("main.go");
    if (await _existFile(file)) {
      return;
    }
    await file.writeAsString('''
package main

import (
	"log"
	"$packageName/cmd"
)

func main() {
	if e := cmd.Execute(); e != nil {
		log.Fatalln(e)
	}
}
''');
  }

  Future<bool> _existFile(File file) async {
    final stat = await file.stat();
    if (stat.type == FileSystemEntityType.notFound) {
      return false;
    }

    if (stat.type == FileSystemEntityType.directory) {
      throw StateError("main.go is a directory");
    }
    return true;
  }
}
